import numpy as np
from .cpu_main import add_dt

def generate_spike_train(N, T, sparsity):
    '''
    Creates an array of spike trains of size (N,T) at a given spike density (sparsity, or total spikes/length of signal)

    returns spike_train (np.array)
    '''
    
    # Check if sparsity is between 0 and 1
    if not (0 <= sparsity <= 1):
        raise ValueError("Spike density must be between 0 and 1.")
    if sparsity == 0:
        return np.zeros((N, T), dtype=int)
    elif sparsity == 1:
        return np.ones((N, T), dtype=int)
        
    num_non_zero = int(N * T * sparsity)
    indices = np.random.choice(N * T, num_non_zero, replace=False)
    spike_train = np.zeros((N, T), dtype=int)
    spike_train.flat[indices] = 1

    return spike_train

def spike_times_to_train(spike_timings, sampling_rate, duration = None):
    """
    Converts an array of spike timings into a spike train (binary array).

    Input:
        spike_timings (numpy.ndarray): Array of spike timings (in seconds).
        sampling_rate (float): Sampling rate of the spike train (in Hz).
        duration (float, optional): Total duration of the spike train (in seconds).
                                    If None, the duration is determined by the maximum spike timing.

    Returns:
        numpy.ndarray: Spike train (binary array) with 1s indicating spike times and 0s elsewhere.
    """

    if not isinstance(spike_timings, (np.ndarray)):
        raise TypeError("spike_timings must be a numpy.ndarray")

    if not isinstance(sampling_rate, (int, float)) or sampling_rate <= 0:
        raise ValueError("sampling_rate must be a positive number")

    if duration is None:
        if not spike_timings: #handle empty list
            duration = 0
        else:
            duration = max(spike_timings) if spike_timings else 0

    if duration < 0:
        raise ValueError("duration must be a non-negative number")

    num_samples = int(duration * sampling_rate)
    spike_train = np.zeros(num_samples, dtype=int)

    for spike_time in spike_timings:
        if 0 <= spike_time <= duration: #check for valid spike times
            sample_index = int(spike_time * sampling_rate)
            if 0 <= sample_index < num_samples: #check for valid sample index
                spike_train[sample_index] = 1

    return spike_train
    

