import numpy as np 
import scipy as sp
from scipy.signal import convolve
from scipy.ndimage import convolve1d

class STTC:
    '''
    Calculate STTC and nulls using numpy backend.
    
    '''
    def __init__(self, dt=0):
        self.dt = dt
    
    def add_dt(self, signal, dt=None):
        '''
        Create tiled signal. Make 0->1 in the surrounding (+/- dt) values around each spike (1)
        This is executed using a 1-d convolution with a rectangle window across each row in the signal.

        For example:
        for dt = 1,
        signal = [1 0 0 0 1 0 0 0 0 1 1 0 0]
        output = [1 1 0 1 1 1 0 0 1 1 1 1 0]

        Inputs
        - signal : (numpy array). Can be 1-d or 2-d array of signals. 
                    Only put in signals in the format (n_cells, n_samples in time)
        - dt : (int, default = None)

        Output
        - convolved_signal (numpy array) in the same type and shape as input.
        '''
        
        if dt is None:
            dt = self.dt
        
        if dt == 0:
            return signal

        change_ndim = False
        if signal.ndim == 1:
            signal = signal.reshape(1, -1)
            change_ndim = True

        # Pad the signal
        padded_signal = np.pad(signal, ((0, 0), (dt, dt)), mode='constant', constant_values=0)

        # Create a window with ones of size (2*dt+1)
        N, T = signal.shape
        window = np.ones(2 * dt + 1)

        # Convolve the padded signal with the window
        convolved_signal = convolve1d(padded_signal, window, axis=1)

        if change_ndim:
            return ((convolved_signal > 0) * 1)[0, dt:-dt].astype(signal.dtype)
        return ((convolved_signal > 0) * 1)[:, dt:-dt].astype(signal.dtype)
    
    def STTC_formula(self, TA, TB, PA, PB):
        '''
        Formula of STTC from the Cutts and Eglen (2014) Paper
        '''
        x = (PA - TB)/(1 - (PA*TB))
        y = (PB - TA)/(1 - (PB*TA))
        STTC = 0.5*(x + y)
        return STTC
    
    def STTC(self, signals, dt=None, ret_all = False):
        '''
        Calculates pairwise STTC of all rows in the signals array.

        Inputs
        - signals : (torch.Tensor). Can be 1-d or 2-d array of signals.
                    Only put in signals in the format (n_cells, n_samples in time)
        - dt : (int, default = None)
        - ret_all : (boolean, default = False) if set to True, returns dictionary with the intermediate calculations of 
                                                TA, TB, PA and PB along with STTC

        Outputs
        - STTC : (torch.Tensor) retruns a (n_cells x n_cells) matrix with each cell giving the paired STTC calculated at set dt.
        - TA, TB, PA, PB : intermediate calculated values that go into the STTC.
        '''
        if dt is None:
            dt = self.dt
            
        N,L = signals.shape
        dt_signals = self.add_dt(signals, dt=dt)
        NB = np.tile(signals.sum(axis=1), (N,1))
        TB = np.tile(dt_signals.mean(axis=1), (N,1))
        TA = TB.T
        
        sparse_signals = sp.sparse.csc_matrix(signals)
        sparse_dt_signals = sp.sparse.csc_matrix(dt_signals)

        DAB = sparse_signals @ sparse_dt_signals.T
        DAB = DAB.toarray()

        PA = DAB/(NB.T)
        PB = DAB.T/(NB)
        
        STTC = self.STTC_formula(TA, TB, PA, PB)
        
        if ret_all:
            return dict(zip(['STTC', 'TA', 'TB', 'PA', 'PB'], [STTC, TA, TB, PA, PB]))
        
        return STTC
    
    def STTC_single(self, A, B, dt=None, ret_all=True):
        '''
        STTC calculation across two input signals without any loops/matrix. I call it the 'Direct Method'
        '''
        if dt is None:
            dt = self.dt
      
        if dt == 0:
            TA, tileA = np.mean(A), A
            TB, tileB = np.mean(B), B
        else:
            tileA = self.add_dt(A, dt)
            TA = np.mean(tileA)

            tileB = self.add_dt(B, dt)
            TB = np.mean(tileB)

        n_spikes_A = np.sum(A)
        n_spikes_B = np.sum(B)

        if n_spikes_A == 0:
            PA = 0
        else:
            PA = np.sum(A.astype(int) & tileB.astype(int))/n_spikes_A
        
        if n_spikes_B == 0:
            PB = 0
        else:
            PB = np.sum(B.astype(int) & tileA.astype(int))/n_spikes_B
        
        STTC = self.STTC_formula(TA, TB, PA, PB)
        if ret_all:
            return dict(zip(['STTC', 'TA', 'TB', 'PA', 'PB'], [STTC, TA, TB, PA, PB]))
        else:
            return STTC
    def shift_signal(self, signals, shift=None, shuffle=False):
        '''
        shuffle or circular shifts signal according to the second dimension
        '''
        if shuffle:
            return signals[:, np.random.permutation(signals.shape[1])]
        return np.roll(signals, shift=shift, axis=1)
        
    def STTC_null(self, signals, n_shifts, dt=None, shuffle=False):
        '''
        Calculates null where each row's (cell) STTC is calculated against a circular-shifted array of every other cell. 
        This is done for (ns = 1 to n_shifts + 1) values
    
        If shuffle=True, instead of circular shifting, the signals are instead randomly shuffled 'n_shift' times.
    
        The output is shaped as (ns, n_cell, n_cell).
        Each cell can be calculated as STTC(A, np.roll(B, shift=ns)) (if shuffle=False)
    
        Inputs
        - signals : (np.array). 2-d array of signals.
                    Only put in signals in the format (n_cells, n_samples in time)
        - n_shifts : (int). Maximal value of shifts to be calculated.
        - dt : (int, default = 0)
        - ret_all : (boolean, default = False) if set to True, returns dictionary with the intermediate calculations of TA, TB, PA and PB along with STTC
        - shuffle : (boolean, default = False) if set to True, shuffles signals instead of circular shifting.
    
        Outputs
        - null : (np.array) retruns a (ns x n_cells x n_cells) matrix with each cell giving the paired STTC calculated at set dt with the j-index being shifted by ns.
    
        '''
        if dt is None:
            dt = self.dt

        if dt == 0:
            return self.STTC_null_nodt(signals=signals, n_shifts=n_shifts, shuffle=shuffle)
        N,L = signals.shape
        null = []

        dt_signals = self.add_dt(signals, dt=dt)
        NB = np.tile(signals.sum(axis=1), (N,1))
        TA = np.tile(dt_signals.mean(axis=1), (N,1)).T


        sparse_signals = sp.sparse.csr_matrix(signals)
        sparse_dt_signals = sp.sparse.csr_matrix(dt_signals)
        
        for ns in range(n_shifts):
            
            shifted = self.shift_signal(signals, shift=ns+1, shuffle=shuffle)
            dt_shifted = self.add_dt(shifted, dt=dt)
            TB = np.tile(dt_shifted.mean(axis=1), (N,1))

            sparse_shifted = sp.sparse.csr_matrix(shifted)
            sparse_dt_shifted = sp.sparse.csr_matrix(dt_shifted)

            DAB = sparse_signals @ sparse_dt_shifted.T
            DBA = sparse_dt_signals @ sparse_shifted.T

            DAB = DAB.toarray()
            DBA = DBA.toarray()
            
            PA = DAB/NB.T
            PB = DBA/NB

            null.append(self.STTC_formula(TA, TB, PA, PB))
            
        null = np.stack(null, axis=0)
        return null

    def STTC_null_nodt(self, signals, n_shifts=100, shuffle=False):
        '''
        Same as STTC.STTC_null() but optimized for when dt = 0.
        '''
        N,L = signals.shape
        
        null = []

        NB = np.tile(signals.sum(axis=1), (N,1))
        TA = np.tile(signals.mean(axis=1), (N,1)).T


        sparse_signals = sp.sparse.csr_matrix(signals)
        for ns in range(n_shifts):
            shifted = self.shift_signal(signals, shift=ns+1, shuffle=shuffle)
            TB = np.tile(shifted.mean(axis=1), (N,1))

            sparse_shifted = sp.sparse.csr_matrix(shifted)
            
            DAB = sparse_signals @ sparse_shifted.T
            DAB = DAB.toarray()
            
            PA = DAB/NB.T
            PB = DAB/NB
            
            null.append(self.STTC_formula(TA, TB, PA, PB))
            
        null = np.stack(null, axis=0)
        
        return null
    
    def z_score(self, signals, dt = None, n_shifts = 100, shuffle=False, ret_all = False):
        '''
        Returns z-scored STTC matrix of input signals by calculating null and using it in the formula :
        z-score = (STTC - null_mean)/null_std

        Inputs
        - signals  : (np.array). 2-d array of signals.
                    Only put in signals in the format (n_cells, n_samples in time)
        - dt       : (int, default = 0) time/frame-lag
        - n_shifts : (int). Maximal value of shifts to be calculated for the null.
        - shuffle  : (boolean, default = False) if set to True, shuffles signals instead of circular shifting.
        - ret_all  : (boolean, default = False) if set to True, returns z-score, STTC and null
    
        Outputs
        - z-score : (np.array) retruns a (n_cells x n_cells) matrix with each cell giving the z-scored value of the STTC metric
        - STTC and null (if ret_all is True)
        
        '''
        if dt is None:
            dt = self.dt

        STTC = self.STTC(signals, dt = dt)

        if dt == 0:
            null = self.STTC_null_nodt(signals=signals, n_shifts=n_shifts, shuffle=shuffle)
        else:
            null = self.STTC_null(signals=signals, dt = dt, n_shifts=n_shifts, shuffle=shuffle)
        if ret_all:
            return (STTC - null.mean(axis=0))/null.std(axis=0), STTC, null    
        return (STTC - null.mean(axis=0))/null.std(axis=0)
            
        
        