import torch

class STTC:
    '''
    Calculate STTC and null using torch Tensors and GPU with cuda backend. 
    
    '''
    def __init__(self, dt=0, device='cuda'):
        '''
        Initialize with device = 'cuda' or specific GPU as required.
        '''
        self.dt = dt
        self.device = device 
        
    
    def add_dt(self, signal, dt=None):
        '''
        Create tiled signal. Make 0->1 in the surrounding (+/- dt) values around each spike (1)
        This is executed using a 1-d convolution with a rectangle window across each row in the signal.

        For example:
        for dt = 1,
        signal = [1 0 0 0 1 0 0 0 0 1 1 0 0]
        output = [1 1 0 1 1 1 0 0 1 1 1 1 0]

        Inputs
        - signal : (torch.Tensor). Can be 1-d or 2-d array of signals. Output is the same type of Tensor as input. 
                    Only put in signals in the format (n_cells, n_samples in time)
        - dt : (int, default = None)

        Output
        - convolved_signal (torch.Tensor) in the same type and shape as input Tensor.
        '''
        if not torch.is_tensor(signal):
            raise ValueError("Input must be a PyTorch tensor.")
        
        if dt is None:
            dt = self.dt
        
        if dt == 0:
            return signal
        
        change_ndim = False
        if signal.ndim==1:
            signal = signal.view(1,-1)
            change_ndim = True


        # Pad the signal
        padded_signal = torch.nn.functional.pad(signal, (dt, dt), value=0)

        # Create a window with ones of size (2*dt+1)
        N,T = signal.shape
        window = torch.ones(N, 2 * dt + 1, device=self.device, dtype=signal.dtype)

        # Convolve the padded signal with the window
        convolved_signal = torch.nn.functional.conv1d(padded_signal, 
                                                      window.unsqueeze(1), padding=dt, groups=N)

        if change_ndim:
            return ((convolved_signal > 0)*1)[0, dt:-dt].type(signal.dtype)
        return ((convolved_signal > 0)*1)[:, dt:-dt].type(signal.dtype)
    
    def STTC_formula(self, TA, TB, PA, PB):
        '''
        Formula of STTC from the Cutts and Eglen (2014) Paper
        '''
        x = (PA - TB)/(1 - (PA*TB))
        y = (PB - TA)/(1 - (PB*TA))
        STTC = 0.5*(x + y)
        return STTC
    
    def STTC(self, signals, dt=None, ret_all = False):
        '''
        Calculates pairwise STTC of all rows in the signals array.

        Inputs
        - signals : (torch.Tensor). Can be 1-d or 2-d array of signals.
                    Only put in signals in the format (n_cells, n_samples in time)
        - dt : (int, default = None)
        - ret_all : (boolean, default = False) if set to True, returns dictionary with the intermediate calculations of 
                                                TA, TB, PA and PB along with STTC

        Outputs
        - STTC : (torch.Tensor) retruns a (n_cells x n_cells) matrix with each cell giving the paired STTC calculated at set dt.
        - TA, TB, PA, PB : intermediate calculated values that go into the STTC.
        '''
        if not torch.is_tensor(signals):
            raise ValueError("Input must be a PyTorch tensor.")
        
        if dt is None:
            dt = self.dt
            
        N,L = signals.shape
        dt_signals = self.add_dt(signals, dt=dt)
        NB = torch.tile(signals.sum(axis=1), (N,1))
        TB = torch.tile(dt_signals.mean(axis=1), (N,1))
        TA = TB.T
        
        DAB = signals @ dt_signals.T

        PA = DAB/(NB.T)
        PB = DAB.T/(NB)
        
        STTC = self.STTC_formula(TA, TB, PA, PB)
        
        if ret_all:
            return dict(zip(['STTC', 'TA', 'TB', 'PA', 'PB'], [STTC, TA, TB, PA, PB]))
        
        return STTC

    def STTC_single(self, A, B, dt=0, ret_all=True):
        '''
        STTC calculation across two input signals without any loops/matrix. I call it the 'Direct Method'
        '''
        if not torch.is_tensor(A) or not torch.is_tensor(B):
            raise ValueError(f"Input must be a PyTorch tensor.  A : {type(A)}, B : {type(B)}")

        if A.ndim>1 or B.ndim>1:
            raise ValueError("Input must be 1-d signals.")

        if dt == 0:
            TA, tileA = A.mean(), A
            TB, tileB = B.mean(), B
        else:
            tileA = self.add_dt(A, dt=dt)
            TA = tileA.mean()

            tileB = self.add_dt(B, dt=dt)
            TB = tileB.mean()

        n_spikes_A = A.sum()
        n_spikes_B = B.sum()

        if n_spikes_A == 0:
            PA = 0
        else:
            PA = torch.matmul(A, tileB).sum()/n_spikes_A

        if n_spikes_B == 0:
            PB = 0
        else:
            PB = torch.matmul(B, tileA).sum()/n_spikes_B

        STTC = self.STTC_formula(TA, TB, PA, PB)
        if ret_all:
            return dict(zip(['STTC', 'TA', 'TB', 'PA', 'PB'], [STTC, TA, TB, PA, PB]))
        else:
            return STTC

    def shift_signal(self, signals, shift=None, shuffle=False):
        '''
        shuffle or circular shifts signal according to the second dimension
        '''
        if not torch.is_tensor(signals):
            raise ValueError("Input must be a PyTorch tensor.")
            
        if shuffle:
            return signals[:, torch.randperm(signals.shape[1])]
        return torch.roll(signals, shifts=shift, dims=1)
        
        
    def STTC_null(self, signals, n_shifts, dt=None, shuffle=False):
        '''
        Calculates null where each row's (cell) STTC is calculated against a circular-shifted array of every other cell. 
        This is done for (ns = 1 to n_shifts + 1) values

        The output is shaped as (ns, n_cell, n_cell).
        Each cell can be calculated as STTC(A, np.roll(B, shifts=ns))

        Inputs
        - signals : (torch.Tensor). Can be 1-d or 2-d array of signals.
                    Only put in signals in the format (n_cells, n_samples in time)
        - n_shifts : (int). Maximal value of shifts to be calculated.
        - dt : (int, default = None)
        - ret_all : (boolean, default = False) if set to True, returns dictionary with the intermediate calculations of TA, TB, PA and PB along with STTC
        - shuffle : (boolean, default = False) if set to True, shuffles signals instead of circular shifting.

        Outputs
        - null : (torch.Tensor) retruns a (ns x n_cells x n_cells) matrix with each cell giving the paired STTC calculated at set dt with the j-index being shifted by ns.

        '''
        if not torch.is_tensor(signals):
            raise ValueError("Input must be a PyTorch tensor.")
        
        if dt is None:
            dt = self.dt

        N,L = signals.shape
        null = []

        dt_signals = self.add_dt(signals, dt=dt)
        NB = torch.tile(signals.sum(axis=1), (N,1))
        TA = torch.tile(dt_signals.mean(axis=1), (N,1)).T

        for ns in range(n_shifts):
            shifted = self.shift_signal(signals, shift=ns+1, shuffle=shuffle)
            dt_shifted = self.add_dt(shifted, dt=dt)
            TB = torch.tile(dt_shifted.mean(axis=1), (N,1))

            DAB = signals @ dt_shifted.T
            DBA = dt_signals @ shifted.T

            PA = DAB/NB.T
            PB = DBA/NB

            null.append(self.STTC_formula(TA, TB, PA, PB))
            
        null = torch.stack(null, dim=0)
        return null
    
    
    
    def z_score(self, signals, n_shifts = 100, dt=None, shuffle=False, ret_all=False):
        '''
        Returns z-scored STTC matrix of input signals by calculating null and using it in the formula :
        z-score = (STTC - null_mean)/null_std

        Inputs
        - signals : (torch.Tensor). Can be 1-d or 2-d array of signals.
                    Only put in signals in the format (n_cells, n_samples in time)
        - n_shifts : (int). Maximal value of shifts to be calculated.
        - dt : (int, default = None)
        - ret_all : (boolean, default = False) if set to True, returns dictionary with the intermediate calculations of TA, TB, PA and PB along with STTC
        - shuffle : (boolean, default = False) if set to True, shuffles signals instead of circular shifting.
    
        Outputs
        - z-score : (torch.Tensor) retruns a (n_cells x n_cells) matrix with each cell giving the z-scored value of the STTC metric
        - STTC and null (if ret_all is True)
        
        '''
        if dt is None:
            dt = self.dt

        STTC = self.STTC(signals, dt = dt)
        null = self.STTC_null(signals, n_shifts=n_shifts, dt=dt)
        if ret_all:
            return (STTC - null.mean(axis=0))/null.std(axis=0), STTC, null    
        return (STTC - null.mean(axis=0))/null.std(axis=0)
        
    
    
        
    
